open Unix
open Pervasives

open Utils

(* port on which the server is listening *)
let port = ref 8080

(* Is there a following proxy *)
let proxy = ref None
let noProxy = ref []

(* filters for pages *)
let forbiddenUrl = ref []
let allowedUrl = ref []

(* filters for headers *)
let forbiddenHeader = ref []

(* fill a list *)
let read_list list name conf =
  let tok = ref (next_token conf) in
  while !tok <> "" && !tok <> "end" do
    list := !tok :: !list;
    tok := next_token conf
  done;
  if !tok = "" then
    (output_string stderr
       ("Invalid configuration file (" ^ name ^ " without end)\n");
     exit 1)

(* read the configuration file *)
let init conf_file =
  (* launch the dns process *)
  Types.from_dns_fds := Dns.init_dns ();
  (* read the conf file *)
  let conf = open_in conf_file in
  let cont = ref true in
  while !cont do
    match next_token conf with
    | "port" ->
        port := int_of_string (next_token conf)
    | "stats" ->
        Types.stats := int_of_string (next_token conf)
    | "debug" ->
        Types.debug := int_of_string (next_token conf)
    | "request" ->
        let req = int_of_string (next_token conf) in
        Types.req_1 := ((req land 1) = 1);
        Types.req_2 := ((req land 2) = 2);
        Types.req_3 := ((req land 4) = 4);
    | "answer" ->
        let ans = int_of_string (next_token conf) in
        Types.ans_1 := ((ans land 1) = 1);
        Types.ans_2 := ((ans land 2) = 2);
        Types.ans_3 := ((ans land 4) = 4);
    | "nextProxy" ->
        let p_host = next_token conf in
        let p_port = int_of_string (next_token conf) in
        let entry = gethostbyname p_host in
        proxy := Some (ADDR_INET (entry.h_addr_list.(0), p_port))
    | "noProxy" ->
        read_list noProxy "noProxy" conf
    | "forbiddenUrl" ->
        read_list forbiddenUrl "forbiddenUrl" conf
    | "allowedUrl" ->
        read_list allowedUrl "allowedUrl" conf
    | "forbiddenHeader" ->
        read_list forbiddenHeader "forbiddenHeader" conf
    | "answerGIF" ->
        let image = next_token conf in
        Types.forbid_url := "HTTP/1.0 200 OK\r\nServer: Proxy http filtrant\r\nAccept-Ranges: bytes\r\nConnection: close\r\nContent-Type: image/gif\r\n\r\n";
        Types.forbid_url := !Types.forbid_url ^ (read_file image)
    | "" -> cont := false
    | s ->
        output_string stderr
          ("Invalid configuration file (unknown token " ^ s ^ ")\n");
        exit 1
  done

(* Does this site need a proxy *)
let need_proxy site =
  let rec need_aux = function
    | [] ->
        !proxy
    | pat::l ->
        if match_pattern site pat
        then None
        else need_aux l
  in
  need_aux !noProxy

(* see if a string matches a pattern in the list *)
let rec match_list str = function
  | [] -> false
  | pat::l ->
      match_pattern str pat || match_list str l

(* test if an url is allowed or not *)
let ok_url url =
  if !allowedUrl = []
  then
    not (match_list url !forbiddenUrl)
  else
    (match_list url !allowedUrl) && not (match_list url !forbiddenUrl)

(* test if this line of header is allowed *)
let ok_header header =
  not (match_list header !forbiddenHeader)
