open Unix
open Pervasives

exception Not_yet
exception Dns_queue of int

let to_dns = ref stdout
let from_dns = ref Unix.stdin

type order =
  | Request of int*string
  | Forget of int

let req_number = ref 0 (* number of the last request *)
let wait_list = ref []
let answers = ref []

(* read an answer from a fds *)
let read_value fds =
  let header = Utils.really_read fds Marshal.header_size in
  let ds = Marshal.data_size header 0 in
  let data = Utils.really_read fds ds in
  Marshal.from_string (header^data) 0

(*************)
(* dns cache *)
(*************)

exception Not_in_cache
type cache_value =
  | Empty
  | Exist of Unix.inet_addr

let cache = Array.create Types.dns_cache_size ("", Empty)
let pos = ref 0 (* pos in cache for next write *)

(* add some value in the cache *)
let add_value host ip =
  cache.(!pos) <- (host, ip);
  pos := (!pos +1) mod Types.dns_cache_size

(* try to find the host in the cache *)
exception Is_in_cache of cache_value

let gethostbycache host =
  let sol =
    (try
      if !pos != 0 then
        (let i = ref (!pos - 1) in
        while !i >= 0 do
          if fst cache.(!i) = host then
            raise (Is_in_cache (snd cache.(!i)))
          else
            decr i
        done);
      if snd cache.(Types.dns_cache_size-1) <> Empty then
        (let i = ref (Types.dns_cache_size-1) in
         while !i >= !pos do
           if fst cache.(!i) = host then
             raise (Is_in_cache (snd cache.(!i)))
           else
             decr i
         done);
      Empty
    with Is_in_cache s -> s) in
  match sol with
  | Empty ->
      raise Not_in_cache
  | Exist ip ->
      ip

(* the gethostbyname is made in a separate process
 * because it can block everything *)
let my_gethostbyname host =
  try
    gethostbycache host
  with Not_in_cache ->
    if !Types.debug > 0 then
      (print_string ("dns call for " ^ host ^ " (");
       print_int (List.length !wait_list);
       print_string ", ";
       print_int (List.length !answers);
       print_string ")\n"; flush stdout);
    incr req_number;
    wait_list := (!req_number, host)::!wait_list;
    output_value !to_dns (Request (!req_number, host));
    flush !to_dns;
    raise (Dns_queue !req_number)


(**************************************)
(* Communication with the dns process *)
(**************************************)

let rec delete number = function
  | [] -> []
  | (nb, _)::l when nb = number -> l
  | a::l -> a::(delete number l)

(* read the next dns answer *)
let dns_answer number =
  let rec find = function
    | [] -> raise Not_yet
    | (nb, ip)::l when nb=number ->
        answers := delete nb !answers;
        (match ip with
        | None ->
            raise Not_found
        | Some ip_addr ->
            ip_addr)
    | a::l -> find l in
  find !answers

(* is this dns call done *)
let is_done number =
  let rec find = function
    | [] -> false
    | (nb, ip)::l when nb=number -> true
    | a::l -> find l in
  find !answers

(* forget this number *)
let forget number =
  wait_list := delete number !wait_list;
  output_value !to_dns (Forget number);
  flush !to_dns

(* read a response *)
let update active_read =
  if List.exists ((=) !Types.from_dns_fds) active_read then
    let nb, ans = read_value !from_dns in
    let rec host_name = function
      | [] -> ()
      | (num, host)::l when nb = num ->
          answers := (num, ans)::!answers;
          (* add the answer in the cache *)
          (match ans with
          | Some ip -> add_value host (Exist ip)
          | _ -> ());
          wait_list := delete nb !wait_list
      | a::l -> host_name l in
    host_name !wait_list


(***************************)
(* the dns child processes *)
(***************************)

(* this process treat one call and die *)
let main_child host outch =
  try
    let entry = gethostbyname host in
    output_value outch (Some entry.h_addr_list.(0));
    flush outch;
  with _ ->
    output_value outch None;
    flush outch

let init_child host =
  let ans_in_fds, ans_out_fds = pipe () in
  let ans_out = out_channel_of_descr ans_out_fds in
  match fork () with
  | 0 -> (* Child process *)
      close ans_in_fds;
      main_child host ans_out;
      exit 0
  | pid -> (* main dns process *)
      close_out ans_out;
      ans_in_fds, pid


(*****************************************)
(*           the dns process             *)
(*****************************************)

(* this will work in a separate process *)
let main_dns in_fds outch =
  (try
    let child_list = ref [] in
    while true do
      let all_fds = in_fds::(List.map (fun (x,y,z) -> x) !child_list) in
      let active_read, _, _ = Unix.select all_fds [] [] (-.1.) in
      let rec update_list = function
        | [] -> []
        | (fds, nb, pid)::l ->
            if List.exists ((=) fds) active_read then
              let ans = read_value fds in
              output_value outch (nb, ans);
              flush outch;
              Unix.close fds;
              ignore (Unix.waitpid [] pid);
              update_list l
            else
              (fds, nb, pid)::(update_list l) in
      child_list := update_list !child_list;
      if List.exists ((=) in_fds) active_read then
        (match read_value in_fds with
        | Request (nb, host) ->
            let fds, pid = init_child host in
            child_list := (fds, nb, pid)::!child_list
        | Forget nb ->
            let rec new_list = function
              | [] -> []
              | (fds, num, pid)::l when num=nb ->
                  Unix.kill pid 9;
                  ignore (Unix.waitpid [] pid);
                  l
              | a::l -> a::(new_list l) in
            child_list := new_list !child_list);
    done
  with _ -> ())

(* Launch the separate process *)
let init_dns () =
  let req_in_fds, req_out_fds = pipe () in
  let ans_in_fds, ans_out_fds = pipe () in
  let req_out = out_channel_of_descr req_out_fds in
  let ans_out = out_channel_of_descr ans_out_fds in
  match fork () with
  | 0 -> (* Child process *)
      close_out req_out;
      close ans_in_fds;
      main_dns req_in_fds ans_out;
      exit 0
  | x -> (* main program *)
      Unix.close req_in_fds;
      close_out ans_out;
      to_dns := req_out;
      from_dns := ans_in_fds;
      ans_in_fds
