open Unix
open Pervasives

open Conf
open Types
open Utils

(**************************************************************)
(* understand *)
(**************************************************************)

(* Try to understand the command line *)
let gere_cmdline cmd conn =
  try
    let deb_url = String.index cmd ' ' in
    let fin_url = String.index_from cmd (deb_url+1) ' ' in
    if (String.sub cmd (deb_url+1) 7) = "http://" then
    (* this line looks good *)
      (let url = String.sub cmd (deb_url+8) (fin_url - deb_url - 8) in
      if ok_url url then
        (* this url is not forbidden *)
        ((match !Types.stats with
        | 0 -> ()
        | 1 ->
            print_string ("ok      http://" ^ url ^ "\n");
            flush stdout
        | _ (* >= 2 *) ->
            print_string ("ok      http://" ^ url ^ " ("
                          ^ (string_of_int (List.length !fread))
                          ^ ")\n"); flush stdout);
         let fin_host = String.index url '/' in
         let total_host = String.sub url 0 fin_host in
         let port, host =
           (try
             let deb_port = 1 + String.index total_host ':' in
             let str_port = String.sub total_host deb_port 
                 ((String.length total_host) - deb_port) in
             let host = String.sub total_host 0 (deb_port - 1) in
             int_of_string str_port, host
           with Not_found ->
            (* no port given, use 80 *)
             80, total_host) in
         (match need_proxy host with
         | None ->
             let deb_ans = String.sub cmd 0 (deb_url+1) in
             let file =
               String.sub url fin_host ((String.length url) - fin_host) in
             let proto = 
               String.sub cmd fin_url ((String.length cmd) - fin_url) in
             let new_cmd_line = deb_ans ^ file ^ proto in
             (try
               let ip_addr =
                 (try
                   Dns.my_gethostbyname host
                 with Not_found ->
                   raise (Unix_error (EUNKNOWNERR 0, "gethostbyname", ""))) in
               let sock = ADDR_INET (ip_addr, port) in
               conn.server <- Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0;
               set_nonblock conn.server;
               conn.state <- CONNECTING;
               fread  := conn.server::!fread;
               (try
                 Unix.connect conn.server sock;
                 conn.state <- ALIVE
               with
               | Unix_error (EINPROGRESS, "connect", "") ->
                   fwrite := conn.server :: !fwrite);
               conn.write_req <- new_cmd_line
             with Dns.Dns_queue nb ->
               conn.state_req <- WAIT_DNS (new_cmd_line, port, nb);
               conn.state <- DNS nb)
         | Some p ->
             conn.server <- Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0;
             set_nonblock conn.server;
             conn.state <- CONNECTING;
             fread  := conn.server::!fread;
             (try
               Unix.connect conn.server p;
               conn.state <- ALIVE
             with
             | Unix_error (EINPROGRESS, "connect", "") ->
                 fwrite := conn.server :: !fwrite);
             conn.write_req <- cmd))
      else
        (* forbidden url *)
        (if !Types.stats > 0 then
          (print_string ("refused http://" ^ url ^ "\n"); flush stdout);
         finish conn !forbid_url))
    else
      (* cannot understand anything *)
      finish conn !invalid_req
  with
  | Not_found
  | Failure "int_of_string" ->
      (* cannot understand anything *)
      finish conn !invalid_req
  | Unix_error (_, "gethostbyname", "") ->
      finish conn !unknown_host
  | Unix_error (_, "connect", "") ->
      finish conn !host_unreach

(* Try to understand what we received *)
let rec compute_read conn =
  match conn.state_req with
  | CMD_LINE ->
      (try
        let pos = String.index conn.read_req '\n' in
        let cmd = String.sub conn.read_req 0 (pos+1) in
        conn.read_req <-
          String.sub conn.read_req (pos+1)
            ((String.length conn.read_req)-pos-1);
        conn.state_req <- HEADERS;
        gere_cmdline cmd conn;
        if !Types.req_1 then
          (print_string conn.write_req; flush stdout);
        match conn.state with
        | DNS _ -> ()
        | _ ->
            compute_read conn
      with Not_found ->
        (* the command line is not finished *)
        ())
  | WAIT_DNS (new_cmd_line, port, nb) ->
      (try
        let ip_addr =
          (* can raise Not_found *)
          Dns.dns_answer nb in
        let sock = ADDR_INET (ip_addr, port) in
        conn.server <- Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0;
        set_nonblock conn.server;
        conn.state <- CONNECTING;
        fread  := conn.server::!fread;
        (try
          Unix.connect conn.server sock;
          conn.state <- ALIVE
        with
        | Unix_error (EINPROGRESS, "connect", "") ->
            fwrite := conn.server :: !fwrite);
        conn.write_req <- new_cmd_line;
        conn.state_req <- HEADERS;
        if !Types.req_1 then
          (print_string conn.write_req; flush stdout);
        compute_read conn
      with
      | Unix_error (_, "connect", "") ->
          finish conn !host_unreach
      | Not_found ->
          (* gethostbyname failed *)
          finish conn !unknown_host
      | Dns.Not_yet ->
          failwith "Internal error in request.ml (Dns.Not_yet)")
  | HEADERS ->
      (try
        let pos = String.index conn.read_req '\n' in
        let header = String.sub conn.read_req 0 (pos+1) in
        conn.read_req <-
          String.sub conn.read_req (pos+1)
            ((String.length conn.read_req)-pos-1);
        if pos <= 2 then
          (* last line of headers *)
          (if !Types.req_2 then
            (print_string header; flush stdout);
           conn.state_req <- CONTENT;
           conn.write_req <- conn.write_req ^ header)
        else
          (if ok_header header then
            (if !Types.req_2 then
              (print_string header; flush stdout);
             conn.write_req <- conn.write_req ^ header));
        compute_read conn
      with Not_found ->
        (* this line of headers is not finished *)
        ())
  | CONTENT ->
      if !Types.req_3 then
        (print_string conn.read_req; flush stdout);
      conn.write_req <- conn.write_req ^ conn.read_req;
      conn.read_req <- ""

(**************************************************************)
(* read and write *)
(**************************************************************)

(* buffer used for read *)
let buf = String.create buf_size

(* Update every connexions *)
let gere_conns active_read active_write conns =
  let rec gere_aux conns = function
    | [] -> conns
    | conn::l ->
        match conn.state with
        | DNS number ->
            (* is the dns call done *)
            if Dns.is_done number then
              (* the answer is arrived ! *)
              compute_read conn;
            (* is there sth to read *)
            if List.exists ((=) conn.client) active_read
                && not (manage_read conn) then
              (* connexion closed *)
              (Dns.forget number;
               close_connexion conn;
               gere_aux conns l)
            else gere_aux (conn::conns) l
        | CONNECTING ->
            let connecting () =
              if List.exists ((=) conn.server) active_write then
                (* We're connected *)
                (conn.state <- ALIVE;
                 manage_write conn conns l)
              else
                (* not connected yet *)
                (fwrite := conn.server::!fwrite;
                 gere_aux (conn::conns) l) in
            if List.exists ((=) conn.client) active_read then
              if manage_read conn then
                (* Nothing more to read now *)
                (if conn.read_req <> "" then compute_read conn;
                 (* see if we can write *)
                 connecting ())
              else
                (* Connexion closed by client *)
                (close_connexion conn;
                 gere_aux conns l)
            else
              (* see if we can write something *)
              connecting ()
        | STARTING
        | ALIVE ->
            if List.exists ((=) conn.client) active_read then
              if manage_read conn then
                (* Nothing more to read now *)
                (if conn.read_req <> "" then compute_read conn;
                (* try to write what must be *)
                 if conn.state = ALIVE then
                   manage_write conn conns l
                 else
                   gere_aux (conn::conns) l)
              else
                (* Connexion closed by client *)
                (close_connexion conn;
                 gere_aux conns l)
            else
              (* try to write what must be *)
              if conn.state = ALIVE then
                manage_write conn conns l
              else
                gere_aux (conn::conns) l
        | FINISHING ->
            gere_aux (conn::conns) l

  (* write what can be and is ready *)
  and manage_write conn conns l =
    (try
      let nb = Unix.write conn.server conn.write_req 0
          (String.length conn.write_req) in
      conn.write_req <- String.sub conn.write_req nb
          ((String.length conn.write_req) - nb);
      if conn.write_req <> "" then
        fwrite := conn.server :: !fwrite;
      gere_aux (conn::conns) l
    with
    | Unix_error (EAGAIN, "write", "") ->
        (* Cannot write any more now *)
        fwrite := conn.server :: !fwrite;
        gere_aux (conn::conns) l
    | Unix_error (_, "write", "")  ->
        (* connexion closed by server *)
        finish conn "";
        gere_aux (conn::conns) l);

  (* read all available data
     * return true if the connexion is still alive
     *        false if the connexion is closed *)
  and manage_read conn =
    let cont = ref true in
    let ok = ref true in
    (try
      while !cont do
        let nb = Unix.read conn.client buf 0 buf_size in
        conn.read_req <- conn.read_req ^ (String.sub buf 0 nb);
        match nb with
        | 0 -> (* connexion closed *)
            cont := false;
            ok := false
        | x when x=buf_size -> (* go on *)
            ()
        | _ -> (* nothing more to read now *)
            cont := false
      done
    with
    | Unix_error (EAGAIN, "read", "") -> (* Nothing more to read now *)
        ()
    | Unix_error (_, "read", "") -> (* connexion closed *)
        ok := false);
    !ok
  in

  gere_aux [] conns
